#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <time.h>
#define DELKA_BUFFERU 16384
#define SOUBOR_SETUP "setup.txt"
#define SOUBOR_DATA "data.txt"
#define SOUBOR_TRENOVANI "trenovani.txt"
#define SOUBOR_VYSTUP "vystup.txt"
#define SOUBOR_KONECNE_PARAMETRY "konecne_parametry.txt"
#define SOUBOR_VAHY_BIN "vahy.bin"
#define real_rand() ((double) rand() / (RAND_MAX + 1.0) - 0.5)

typedef struct {
  unsigned radek, sloupek;
} SOURADNICE_SOM;

unsigned cti_unsigned_ze_souboru(FILE *fr)
{
  int c;
  unsigned cislo;
  if (fscanf(fr, "%u", &cislo) != 1)
    printf ("Chybi hodnota v souboru!\n");
  printf("Precteno %u", cislo);
  while (((c = getc(fr)) != '\n') && !feof(fr))
    putchar(c);
  putchar('\n');
  return cislo;
}

unsigned zjisti_pocet_radku_dat(char *jmeno_souboru)
{
  FILE *fr;
  char buffer[DELKA_BUFFERU];
  unsigned pocet_radku = 0;
  double cislo;
  if ((fr = fopen(jmeno_souboru, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    exit(1);
  }
  while (fgets(buffer, DELKA_BUFFERU, fr)) {
    if (sscanf(buffer, "%lf", &cislo) == 1) {
      pocet_radku++;
    }
    buffer[0] = '\0';
  }
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    exit(1);
  }
  return pocet_radku;
}

void zaloz_matici_double(unsigned delka_vzoru, unsigned radek, double ***x, double **p_x)
{
  unsigned i;
  if ((*p_x = (double *) malloc(radek * delka_vzoru * sizeof(double))) == NULL) {
    printf("Malo pameti!\n");
    exit(1);
  }
  if ((*x = (double **) malloc(radek * sizeof(double *))) == NULL) {
    printf("Malo pameti!\n");
    exit(1);
  }
  for (i = 0; i < radek; i++) {
    (*x)[i] = *p_x + i * delka_vzoru;
  }
}

void uvolni_pamet_pro_matici_double(double ***x, double **p_x)
{
  free((void *) *p_x);
  *p_x = NULL;
  free((void *) *x);
  *x = NULL;
}

void zaloz_kostku_double(unsigned radek, unsigned sloupek, unsigned delka_vzoru, double ****x, double ***p_p_x, double **p_x)
{
  unsigned i;
  if ((*p_x = (double *) malloc(radek * sloupek * delka_vzoru * sizeof(double))) == NULL) {
    printf("Malo pameti!\n");
    exit(1);
  }
  if ((*p_p_x = (double **) malloc(radek * sloupek * sizeof(double *))) == NULL) {
    printf("Malo pameti!\n");
    exit(1);
  }
  if ((*x = (double ***) malloc(radek * sizeof(double **))) == NULL) {
    printf("Malo pameti!\n");
    exit(1);
  }
  for (i = 0; i < radek * sloupek; i++) {
    (*p_p_x)[i] = *p_x + i * delka_vzoru;
  }
  for (i = 0; i < radek; i++) {
    (*x)[i] = *p_p_x + i * sloupek;
  }
}

void uvolni_pamet_pro_kostku_double(double ****x, double ***p_p_x, double **p_x)
{
  free((void *) *p_x);
  *p_x = NULL;
  free((void *) *p_p_x);
  *p_p_x = NULL;
  free((void *) *x);
  *x = NULL;
}

void nacti_best_pred(unsigned pocet_radku_dat, SOURADNICE_SOM *best_pred)
{
  FILE *fr;
  unsigned k;
  if ((fr = fopen(SOUBOR_VYSTUP, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_VYSTUP);
    exit(1);
  }
  for (k = 0; k < pocet_radku_dat; k++) {
    if (fscanf(fr, "%u\t%u\n", &(best_pred[k].radek), &(best_pred[k].sloupek)) != 2) {
      printf("Chybi hodnota v souboru %s!", SOUBOR_VYSTUP);
      exit(1);
    }
  }
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_VYSTUP);
    exit(1);
  }
}

void nacti_vahy_SOM(FILE *fr, unsigned vyska, unsigned sirka, unsigned delka_vzoru, double ***SOM)
{
  int i, j, k;
  for (i = 0; i < vyska; i++) {
    for (j = 0; j < sirka; j++) {
      for (k = 0; k < delka_vzoru; k++) {
        if (fread((void *) &(SOM[i][j][k]), sizeof(double), 1, fr) == 0) {
          printf("Nepodarila se nacist data!\n");
          exit(1);
        }
      }
    }
  }
}

void inicializuj_vahy_SOM(unsigned vyska, unsigned sirka, unsigned delka_vzoru, double ***SOM, unsigned pocet_radku_dat, SOURADNICE_SOM *best_pred)
{
  FILE *fr;
  unsigned i, j, k;
  if ((fr = fopen(SOUBOR_VAHY_BIN, "rb")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit. ", SOUBOR_VAHY_BIN);
    printf("Sit se bude ucit z nahodnych vah.\n");
    for (i = 0; i < vyska; i++) {
      for (j = 0; j < sirka; j++) {
        for (k = 0; k < delka_vzoru; k++) {
          SOM[i][j][k] = real_rand();
        }
      }
    }
  } else {
      nacti_vahy_SOM(fr, vyska, sirka, delka_vzoru, SOM);
      if (fclose(fr) == EOF) {
        printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_VAHY_BIN);
        exit(1);
      }
      nacti_best_pred(pocet_radku_dat, best_pred);
    }
}

void vypis_vahy_SOM(unsigned vyska, unsigned sirka, unsigned delka_vzoru, double ***SOM)
{
  FILE *fw;
  int i, j, k;
  if ((fw = fopen(SOUBOR_VAHY_BIN, "wb")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_VAHY_BIN);
    exit(1);
  }
  for (i = 0; i < vyska; i++) {
    for (j = 0; j < sirka; j++) {
      for (k = 0; k < delka_vzoru; k++) {
        if ((fwrite((void *) &(SOM[i][j][k]), sizeof(double), 1, fw)) != 1) {
          printf("Nepodarila se zapsat data!\n");
          exit(1);
        }
      }
    }
  }
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_VAHY_BIN);
    exit(1);
  }
}

void nacti_data(unsigned delka_vzoru, unsigned pocet_radku_dat, double **vstup)
{
  FILE *fr;
  unsigned i, j;
  if ((fr = fopen(SOUBOR_DATA, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_DATA);
    exit(1);
  }
  for (i = 0; i < pocet_radku_dat; i++) {
    for (j = 0; j < delka_vzoru; j++) {
      if (fscanf(fr, "%lf", &vstup[i][j]) != 1) {
        printf("Chybi hodnota v souboru %s!\n", SOUBOR_DATA);
      }
    }
  }
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_DATA);
    exit(1);
  }
}

double vzdalenost_vektor_SOM(unsigned delka_vzoru, double *radek_vstupu, double *SOM)
{
  unsigned l;
  double vzdalenost = 0;
  for (l = 0; l < delka_vzoru; l++) {
    vzdalenost += (radek_vstupu[l] - SOM[l]) * (radek_vstupu[l] - SOM[l]);
  }
  return vzdalenost;
}

SOURADNICE_SOM get_winning_neuron(unsigned vyska, unsigned sirka, unsigned delka_vzoru, double *radek_vstupu, double ***SOM)
{
  SOURADNICE_SOM winner;
  unsigned i, j;
  double vzdalenost, min_vzdalenost = DBL_MAX;
  for (i = 0; i < vyska; i++) {
    for (j = 0; j < sirka; j++) {
      vzdalenost = vzdalenost_vektor_SOM(delka_vzoru, radek_vstupu, SOM[i][j]);
      if (vzdalenost < min_vzdalenost) {
        min_vzdalenost = vzdalenost;
        winner.radek = i;
        winner.sloupek = j;
      }
    }
  }
  return winner;
}

void trenuj(FILE *fw, unsigned vyska, unsigned sirka, unsigned delka_vzoru, unsigned pocet_radku_dat, double **vstup, double ***SOM,
            unsigned *p_krok, unsigned pokles_radius, unsigned pokles_uprava_vah, unsigned min_radius, unsigned *rozdil, SOURADNICE_SOM *best_pred)
{
  const unsigned MAX_VZDALENOST = vyska * vyska + sirka * sirka;
  SOURADNICE_SOM winning_neuron;
  unsigned i, j, k, l, vzdalenost;
  double eta_radius, eta_uprava_vah, radius;
  do {
    *rozdil = 0;
    (*p_krok)++;
    eta_radius = 1.0 / (1.0 + (double) (*p_krok) / pokles_radius);  /* 1.0 / (*p_krok); */
    eta_uprava_vah = 1.0 / (1.0 + (double) (*p_krok) / pokles_uprava_vah);  /* 1.0 / (*p_krok); */
    radius = MAX_VZDALENOST * eta_radius;
    for (k = 0; k < pocet_radku_dat; k++) {
      winning_neuron = get_winning_neuron(vyska, sirka, delka_vzoru, vstup[k], SOM);
      for (i = 0; i < vyska; i++) {
        for (j = 0; j < sirka; j++) {
          vzdalenost = (winning_neuron.radek - i) * (winning_neuron.radek - i) + (winning_neuron.sloupek - j) * (winning_neuron.sloupek - j);
          if (vzdalenost < radius) {
            for (l = 0; l < delka_vzoru; l++) {
              SOM[i][j][l] += eta_uprava_vah * (vstup[k][l] - SOM[i][j][l]);
            }
          }
        }
      }
      *rozdil += (best_pred[k].radek - winning_neuron.radek) * (best_pred[k].radek - winning_neuron.radek) +
                 (best_pred[k].sloupek - winning_neuron.sloupek) * (best_pred[k].sloupek - winning_neuron.sloupek);
      best_pred[k].radek = winning_neuron.radek;
      best_pred[k].sloupek = winning_neuron.sloupek;
    }
    fprintf(fw, "%u\t%f\n", *rozdil, radius);
    fflush(fw);
  } while (radius > min_radius);
}

void vypis_data_na_SOM(unsigned pocet_radku_dat, SOURADNICE_SOM *best_pred, unsigned krok, unsigned cas, unsigned rozdil)
{
  FILE *fw;
  unsigned k;
  if ((fw = fopen(SOUBOR_VYSTUP, "w")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_VYSTUP);
    exit(1);
  }
  for (k = 0; k < pocet_radku_dat; k++) {
    fprintf(fw, "%u\t%u\n", best_pred[k].radek, best_pred[k].sloupek);
  }
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_VYSTUP);
    exit(1);
  }
  if ((fw = fopen(SOUBOR_KONECNE_PARAMETRY, "w")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_KONECNE_PARAMETRY);
    exit(1);
  }
  fprintf(fw, "%u\tKrok\n", krok);
  fprintf(fw, "%u\tSeed for random number generator\n", cas);
  fprintf(fw, "%u\tRozdil\n", rozdil);
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_KONECNE_PARAMETRY);
    exit(1);
  }
}

int main()
{
  FILE *fr, *fw;
  SOURADNICE_SOM *best_pred;
  unsigned cas, vyska_rastru, sirka_rastru, delka_vzoru, pocet_radku_dat, krok, pokles_radius, pokles_uprava_vah, min_radius, rozdil;
  double **vstup, *p_vstup, ***SOM, **p_columns, *p_data;
  if ((fr = fopen(SOUBOR_SETUP, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_SETUP);
    return 1;
  }
  vyska_rastru = cti_unsigned_ze_souboru(fr);
  sirka_rastru = cti_unsigned_ze_souboru(fr);
  delka_vzoru = cti_unsigned_ze_souboru(fr);
  pokles_radius = cti_unsigned_ze_souboru(fr);
  pokles_uprava_vah = cti_unsigned_ze_souboru(fr);
  min_radius = cti_unsigned_ze_souboru(fr);
  krok = cti_unsigned_ze_souboru(fr);
  cas = cti_unsigned_ze_souboru(fr);
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_SETUP);
    return 1;
  }
  if (cas == 0) {
    cas = time(NULL);
  }
  srand(cas);
  pocet_radku_dat = zjisti_pocet_radku_dat(SOUBOR_DATA);
  zaloz_matici_double(delka_vzoru, pocet_radku_dat, &vstup, &p_vstup);
  zaloz_kostku_double(vyska_rastru, sirka_rastru, delka_vzoru, &SOM, &p_columns, &p_data);
  if ((best_pred = (SOURADNICE_SOM *) calloc(pocet_radku_dat, sizeof(SOURADNICE_SOM))) == NULL) {
    printf("Malo pameti!\n");
    return 1;
  }
  inicializuj_vahy_SOM(vyska_rastru, sirka_rastru, delka_vzoru, SOM, pocet_radku_dat, best_pred);
  nacti_data(delka_vzoru, pocet_radku_dat, vstup);
  if ((fw = fopen(SOUBOR_TRENOVANI, "w")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_TRENOVANI);
    return 1;
  }
  fprintf(fw, "Rozdil\tRadius\n");
  trenuj(fw, vyska_rastru, sirka_rastru, delka_vzoru, pocet_radku_dat, vstup, SOM, &krok, pokles_radius, pokles_uprava_vah, min_radius, &rozdil, best_pred);
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_TRENOVANI);
    return 1;
  }
  vypis_data_na_SOM(pocet_radku_dat, best_pred, krok, cas, rozdil);
  vypis_vahy_SOM(vyska_rastru, sirka_rastru, delka_vzoru, SOM);
  free((void *) best_pred);
  best_pred = NULL;
  uvolni_pamet_pro_matici_double(&vstup, &p_vstup);
  uvolni_pamet_pro_kostku_double(&SOM, &p_columns, &p_data);
  printf("Hello world!\n");
  return 0;
}
/*Program Kohonenova_sit;
Uses crt, graph;
Const delka_vzoru = 2;
      r = 100;
      s = 100;
      pokles = 20; {Po kolika krocch klesne eta na polovinu. m je to vt, tm vce to kategorizuje.}
Type Vzor = array [1..delka_vzoru] of Real;
     Uk = ^Vzor;
Var w : array [0..r, 0..s] of Uk;
    i, j, k, l, best_i, best_j, GrDriver, GrMode, ErrCode : Integer;
    krok : LongInt;
    vzdalenost, min_vzdalenost, max_vzdalenost, radius, eta : Real;
    x : Vzor;
    c : Char;
    f, fv : Text;
Begin GrDriver := detect;
      InitGraph (GrDriver, GrMode,'');
      ErrCode := GraphResult;
      if ErrCode = GrOK then
      begin Randomize; krok := 0; max_vzdalenost := r * r + s * s;
            for i := 0 to r do for j := 0 to s do new (w [i,j]);
            for i := 0 to r do for j := 0 to s do for k := 1 to delka_vzoru do
            w [i,j]^[k] := integer (random (2000)) / 1000 - 1;
            assign (f, 'data.txt'); assign (fv, 'vysledek.txt');
            repeat ClearViewPort; reset (f); l := 0;
                   krok := krok + 1;
                   eta := 1 / (1 + krok / pokles)  {1/krok};
                   radius := max_vzdalenost * eta;
                   while not SeekEOF (f) do
                   begin for k := 1 to delka_vzoru do read (f, x[k]);
                         min_vzdalenost := MaxLongInt;
                         for i := 0 to r do for j := 0 to s do
                         begin vzdalenost := 0;
                               for k := 1 to delka_vzoru do
                               vzdalenost := vzdalenost + sqr (x[k] - w[i,j]^[k]);
                               if vzdalenost < min_vzdalenost then
                               begin min_vzdalenost := vzdalenost;
                                     best_i := i; best_j := j
                               end;
                         end;
                         for i := 0 to r do for j := 0 to s do
                         begin vzdalenost := sqr (best_i - i) + sqr (best_j - j);
                               if vzdalenost < radius then
                               for k := 1 to delka_vzoru do
                               w[i,j]^[k] := w[i,j]^[k] + eta * (x[k] - w[i,j]^[k]);
                         end;
                         l := l + 1; PutPixel (best_i, best_j, l mod 15)
                   end;
            until KeyPressed; c := ReadKey;
            rewrite (fv); reset (f); l := 0;
            while not SeekEOF (f) do
            begin for k := 1 to delka_vzoru do read (f, x[k]);
                  min_vzdalenost := MaxLongInt;
                  for i := 0 to r do for j := 0 to s do
                  begin vzdalenost := 0;
                        for k := 1 to delka_vzoru do
                        vzdalenost := vzdalenost + sqr (x[k] - w[i,j]^[k]);
                        if vzdalenost < min_vzdalenost then
                        begin min_vzdalenost := vzdalenost;
                              best_i := i; best_j := j
                        end;
                  end;
                  l := l + 1; WriteLn (fv, l, #9, best_i, #9, best_j)
            end;
            WriteLn (fv, krok - 1); close (f); close (fv)
      end
      else writeln('Chyba grafiky: ', GraphErrorMsg(ErrCode));
      for i := 0 to r do for j := 0 to s do dispose (w [i,j]);
      c := ReadKey;
end.*/
